##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Demonstration of HDA flowsheet with optimization.

"""

# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           ConcreteModel,
                           Expression,
                           Objective,
                           minimize,
                           SolverFactory,
                           TransformationFactory,
                           value,
                           Var)
from pyomo.network import Arc, SequentialDecomposition

# Import IDAES core
from idaes.core import FlowsheetBlock

# Import Unit Model Modules
from . import hda_ideal_VLE as thermo_props
from . import hda_reaction as reaction_props

# Import Unit Model Modules
from idaes.unit_models import (PressureChanger,
                               StoichiometricReactor,
                               Flash,
                               Heater,
                               Mixer,
                               Separator as Splitter)

from idaes.unit_models.pressure_changer import ThermodynamicAssumption

from idaes.core.util.model_statistics import degrees_of_freedom


def main():
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})

    # Add property packages to flowsheet library
    m.fs.thermo_params = thermo_props.HDAParameterBlock()
    m.fs.reaction_params = reaction_props.HDAReactionParameterBlock(
        default={"property_package": m.fs.thermo_params})

    # Create unit models
    m.fs.M101 = Mixer(default={"property_package": m.fs.thermo_params,
                               "inlet_list": ["toluene_feed", "hydrogen_feed",
                                              "vapor_recycle"]})

    m.fs.H101 = Heater(default={"property_package": m.fs.thermo_params,
                                "has_pressure_change": False,
                                "has_phase_equilibrium": True})

    m.fs.R101 = StoichiometricReactor(
            default={"property_package": m.fs.thermo_params,
                     "reaction_package": m.fs.reaction_params,
                     "has_heat_of_reaction": True,
                     "has_heat_transfer": True,
                     "has_pressure_change": False})

    m.fs.F101 = Flash(default={"property_package": m.fs.thermo_params,
                               "has_heat_transfer": True,
                               "has_pressure_change": True})

    m.fs.S101 = Splitter(default={"property_package": m.fs.thermo_params,
                                  "ideal_separation": False,
                                  "outlet_list": ["purge", "recycle"]})

    # This is needed to avoid pressure degeneracy in recylce loop
    m.fs.C101 = PressureChanger(default={
            "property_package": m.fs.thermo_params,
            "compressor": True,
            "thermodynamic_assumption": ThermodynamicAssumption.isothermal})

    m.fs.F102 = Flash(default={"property_package": m.fs.thermo_params,
                               "has_heat_transfer": True,
                               "has_pressure_change": True})

    m.fs.H101.control_volume.scaling_factor_energy = 1e-3
    m.fs.R101.control_volume.scaling_factor_energy = 1e-3
    m.fs.F101.control_volume.scaling_factor_energy = 1e-3
    m.fs.C101.control_volume.scaling_factor_energy = 1e-3
    m.fs.F102.control_volume.scaling_factor_energy = 1e-3

    # Connect units
    m.fs.s03 = Arc(source=m.fs.M101.outlet, destination=m.fs.H101.inlet)
    m.fs.s04 = Arc(source=m.fs.H101.outlet, destination=m.fs.R101.inlet)
    m.fs.s05 = Arc(source=m.fs.R101.outlet, destination=m.fs.F101.inlet)
    m.fs.s06 = Arc(source=m.fs.F101.vap_outlet, destination=m.fs.S101.inlet)
    m.fs.s08 = Arc(source=m.fs.S101.recycle, destination=m.fs.C101.inlet)
    m.fs.s09 = Arc(source=m.fs.C101.outlet,
                   destination=m.fs.M101.vapor_recycle)
    m.fs.s10 = Arc(source=m.fs.F101.liq_outlet, destination=m.fs.F102.inlet)

    TransformationFactory("network.expand_arcs").apply_to(m)

    # Set operating conditions
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Vap", "benzene"].fix(1e-5)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Vap", "toluene"].fix(1e-5)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Vap", "hydrogen"].fix(1e-5)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Vap", "methane"].fix(1e-5)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Liq", "benzene"].fix(1e-5)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Liq", "toluene"].fix(0.30)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Liq", "hydrogen"].fix(1e-5)
    m.fs.M101.toluene_feed.flow_mol_phase_comp[0, "Liq", "methane"].fix(1e-5)
    m.fs.M101.toluene_feed.temperature.fix(303.2)
    m.fs.M101.toluene_feed.pressure.fix(350000)

    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Vap", "benzene"].fix(1e-5)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Vap", "toluene"].fix(1e-5)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Vap", "hydrogen"].fix(0.30)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Vap", "methane"].fix(0.02)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Liq", "benzene"].fix(1e-5)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Liq", "toluene"].fix(1e-5)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Liq", "hydrogen"].fix(1e-5)
    m.fs.M101.hydrogen_feed.flow_mol_phase_comp[0, "Liq", "methane"].fix(1e-5)
    m.fs.M101.hydrogen_feed.temperature.fix(303.2)
    m.fs.M101.hydrogen_feed.pressure.fix(350000)

    m.fs.H101.outlet.temperature.fix(600)

    m.fs.R101.conversion = Var(initialize=0.75, bounds=(0, 1))

    m.fs.R101.conv_constraint = Constraint(
        expr=m.fs.R101.conversion * m.fs.R101.inlet.
        flow_mol_phase_comp[0, "Vap", "toluene"] ==
        (m.fs.R101.inlet.flow_mol_phase_comp[0, "Vap", "toluene"] -
         m.fs.R101.outlet.flow_mol_phase_comp[0, "Vap", "toluene"]))

    m.fs.R101.conversion.fix(0.75)
    m.fs.R101.heat_duty.fix(0)

    m.fs.F101.vap_outlet.temperature.fix(325.0)
    m.fs.F101.deltaP.fix(0)

    m.fs.S101.split_fraction[0, "purge"].fix(0.2)

    m.fs.C101.outlet.pressure.fix(350000)

    m.fs.F102.vap_outlet.temperature.fix(375)
    m.fs.F102.deltaP.fix(-200000)

    # Define expressions
    # Product purity
    m.fs.purity = Expression(
        expr=m.fs.F102.vap_outlet.flow_mol_phase_comp[0, "Vap", "benzene"] /
        (m.fs.F102.vap_outlet.flow_mol_phase_comp[0, "Vap", "benzene"]
         + m.fs.F102.vap_outlet.flow_mol_phase_comp[0, "Vap", "toluene"]))

    # Operating cost ($/yr)
    m.fs.cooling_cost = Expression(expr=0.212e-7 * -m.fs.F101.heat_duty[0] +
                                   0.212e-7 * -m.fs.R101.heat_duty[0])
    m.fs.heating_cost = Expression(expr=2.2e-7 * m.fs.H101.heat_duty[0] +
                                   1.9e-7 * m.fs.F102.heat_duty[0])
    m.fs.operating_cost = Expression(expr=(3600 * 24 * 365 *
                                           (m.fs.heating_cost +
                                            m.fs.cooling_cost)))
    print(degrees_of_freedom(m))

    # Initialize Units
    # Define method for initialising each block
    def function(unit):
        unit.initialize(outlvl=1)

    # Create instance of sequential decomposition tool
    seq = SequentialDecomposition()
    seq.options.select_tear_method = "heuristic"
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = 5

    # Determine tear stream and calculation order
    G = seq.create_graph(m)
    heu_result = seq.tear_set_arcs(G, method="heuristic")
    order = seq.calculation_order(G)

    # Display tear stream and calculation order
    for o in heu_result:
        print(o.name)
    for o in order:
        for oo in o:
            print(oo.name)

    # Set guesses for tear stream
    tear_guesses = {
        "flow_mol_phase_comp": {
                (0, "Vap", "benzene"): 1e-5,
                (0, "Vap", "toluene"): 1e-5,
                (0, "Vap", "hydrogen"): 0.30,
                (0, "Vap", "methane"): 0.02,
                (0, "Liq", "benzene"): 1e-5,
                (0, "Liq", "toluene"): 0.30,
                (0, "Liq", "hydrogen"): 1e-5,
                (0, "Liq", "methane"): 1e-5},
        "temperature": {0: 303},
        "pressure": {0: 350000}}
    seq.set_guesses_for(m.fs.H101.inlet, tear_guesses)

    # Run sequential initialization
    seq.run(m, function)

    # # Create a solver
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6}
    solver.options = {'tol': 1e-6, 'max_iter': 5000}
    results = solver.solve(m, tee=True)

    # Print results
    print("M101 Outlet")
    m.fs.M101.outlet.display()

    print("H101 Outlet")
    m.fs.H101.outlet.display()

    print("R101 Outlet")
    m.fs.R101.outlet.display()

    print("F101")
    m.fs.F101.liq_outlet.display()
    m.fs.F101.vap_outlet.display()

    print("F102")
    m.fs.F102.liq_outlet.display()
    m.fs.F102.vap_outlet.display()

    print("Purge")
    m.fs.S101.purge.display()

    print("Purity:", value(m.fs.purity))

    # Optimize process
    m.fs.objective = Objective(sense=minimize, expr=m.fs.operating_cost)

    # Decision variables
    m.fs.H101.outlet.temperature.unfix()
    m.fs.R101.heat_duty.unfix()
    m.fs.F101.vap_outlet.temperature.unfix()
    m.fs.F102.vap_outlet.temperature.unfix()
    m.fs.F102.deltaP.unfix()

    # Variable bounds
    m.fs.H101.outlet.temperature[0].setlb(500)
    m.fs.H101.outlet.temperature[0].setub(600)
    m.fs.R101.outlet.temperature[0].setlb(600)
    m.fs.R101.outlet.temperature[0].setub(800)
    m.fs.F101.vap_outlet.temperature[0].setlb(298.0)
    m.fs.F101.vap_outlet.temperature[0].setub(450.0)
    m.fs.F102.vap_outlet.temperature[0].setlb(298.0)
    m.fs.F102.vap_outlet.temperature[0].setub(450.0)
    m.fs.F102.vap_outlet.pressure[0].setlb(105000)
    m.fs.F102.vap_outlet.pressure[0].setub(110000)

    # Additional Constraints
    m.fs.overhead_loss = Constraint(
        expr=m.fs.F101.vap_outlet.flow_mol_phase_comp[0, "Vap", "benzene"] <=
        0.20 * m.fs.R101.outlet.flow_mol_phase_comp[0, "Vap", "benzene"])

    m.fs.product_flow = Constraint(
        expr=m.fs.F102.vap_outlet.flow_mol_phase_comp[0, "Vap", "benzene"] >=
        0.15)

    m.fs.product_purity = Constraint(expr=m.fs.purity >= 0.80)

    # Create a solver
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6}
    solver.options = {'tol': 1e-6, 'max_iter': 5000}
    results = solver.solve(m, tee=True)

    # Print optimization results
    print()
    print("Optimal Solution")
    m.fs.operating_cost.display()
    m.fs.H101.heat_duty.display()
    m.fs.R101.heat_duty.display()
    m.fs.F101.heat_duty.display()
    m.fs.F102.heat_duty.display()

    # Print results
    print("M101 Outlet")
    m.fs.M101.outlet.display()

    print("H101 Outlet")
    m.fs.H101.outlet.display()

    print("R101 Outlet")
    m.fs.R101.outlet.display()

    print("F101")
    m.fs.F101.liq_outlet.display()
    m.fs.F101.vap_outlet.display()

    print("F102")
    m.fs.F102.liq_outlet.display()
    m.fs.F102.vap_outlet.display()

    print("Purge")
    m.fs.S101.purge.display()

    print("Recycle")
    m.fs.S101.recycle.display()

    print("Purity:", value(m.fs.purity))

    # For testing purposes
    return(m, results)


if __name__ == "__main__":
    main()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Basic PID control test using a tank with steam flowing through it.  The
controller is set up to maintain a tank pressure by adjusting the inlet valve
position.

           valve_1   +----+
  steam ----|><|-->--| ta |    valve_2
                     | nk |-----|><|--->--- steam
                     +----+

To test there are two basic things:
  1. That the problem with control goes to the steady state solution
  2. That the dynamic problem with control can be split smoothly across two
     different models representing different adjacent time periods.
"""

__author__ = "John Eslick"
import pytest
import pyomo.environ as pyo
from pyomo.network import Arc
from idaes.unit_models.power_generation import SteamValve
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.unit_models import Heater
from idaes.property_models import iapws95
from idaes.core.util import copy_port_values as _set_port
from idaes.core.util.plot import stitch_dynamic
from idaes.dynamic import PIDBlock

solver_available = pyo.SolverFactory('ipopt').available()
prop_available = iapws95.iapws95_available()

def _add_inlet_pressure_step(m, time=1, value=6.0e5):
    """Easy function to add an inlet pressure step change

    Args:
        m (ConcreteModel): A valve and tank model
        time (float): Time for the step to occur
        value (float): New pressure at time

    Returns:
        None
    """
    for t in m.fs.time:
        if t >= time:
            m.fs.valve_1.inlet.pressure[t].fix(value)

def create_model(steady_state=True, time_set=[0,3], nfe=5, calc_integ=True):
    """ Create a test model and solver

    Args:
        steady_state (bool): If True, create a steady state model, otherwise
            create a dynamic model
        time_set (list): The begining and end point of the time domain
        nfe (int): Number of finite elements argument for the DAE transformation
        calc_integ (bool): If Ture, calculate in the initial condition for
            the integral term, else use a fixed variable (fs.ctrl.err_i0), flase
            is the better option if you have a value from a previous time period

    Returns
        (tuple): (ConcreteModel, Solver)
    """
    if steady_state:
        fs_cfg = {"dynamic":False}
        model_name = "Steam Tank, Steady State"
    else:
        fs_cfg = {"dynamic":True, "time_set":time_set}
        model_name = "Steam Tank, Dynamic"
    m = pyo.ConcreteModel(name=model_name)
    m.fs = FlowsheetBlock(default=fs_cfg)
    # Create a property parameter block
    m.fs.prop_water = iapws95.Iapws95ParameterBlock(
        default={"phase_presentation":iapws95.PhaseType.LG})
    # Create the valve and tank models
    m.fs.valve_1 = SteamValve(default={
        "dynamic":False,
        "has_holdup":False,
        "material_balance_type":MaterialBalanceType.componentTotal,
        "property_package":m.fs.prop_water})
    m.fs.tank = Heater(default={
        "has_holdup":True,
        "material_balance_type":MaterialBalanceType.componentTotal,
        "property_package":m.fs.prop_water})
    m.fs.valve_2 = SteamValve(default={
        "dynamic":False,
        "has_holdup":False,
        "material_balance_type":MaterialBalanceType.componentTotal,
        "property_package":m.fs.prop_water})
    # Connect the models
    m.fs.v1_to_t = Arc(source=m.fs.valve_1.outlet, destination=m.fs.tank.inlet)
    m.fs.t_to_v2 = Arc(source=m.fs.tank.outlet, destination=m.fs.valve_2.inlet)
    # The control volume block doesn't assume the two phases are in equilibrium
    # by default, so I'll make that assumption here, I don't actually expect
    # liquid to form but who knows. The phase_fraction in the control volume is
    # volumetric phase fraction hense the densities.
    @m.fs.tank.Constraint(m.fs.time)
    def vol_frac_vap(b, t):
        return b.control_volume.properties_out[t].phase_frac["Vap"]\
            *b.control_volume.properties_out[t].dens_mol\
            /b.control_volume.properties_out[t].dens_mol_phase["Vap"]\
            == b.control_volume.phase_fraction[t, "Vap"]
    # Add the stream constraints and do the DAE transformation
    pyo.TransformationFactory('network.expand_arcs').apply_to(m.fs)
    if not steady_state:
        pyo.TransformationFactory('dae.finite_difference').apply_to(
            m.fs, nfe=nfe, wrt=m.fs.time, scheme='BACKWARD')
    # Fix the derivative variables to zero at time 0 (steady state assumption)
    m.fs.fix_initial_conditions()
    # A tank pressure reference that's directly time-indexed
    m.fs.tank_pressure = pyo.Reference(
        m.fs.tank.control_volume.properties_out[:].pressure)
    # Add a controller
    m.fs.ctrl = PIDBlock(default={"pv":m.fs.tank_pressure,
                                  "output":m.fs.valve_1.valve_opening,
                                  "upper":1.0,
                                  "lower":0.0,
                                  "calculate_initial_integral":calc_integ})
    m.fs.ctrl.deactivate() # Don't want controller turned on by default
    # Fix the input variables
    m.fs.valve_1.inlet.enth_mol.fix(50000)
    m.fs.valve_1.inlet.pressure.fix(5e5)
    m.fs.valve_2.outlet.pressure.fix(101325)
    m.fs.valve_1.Cv.fix(0.001)
    m.fs.valve_2.Cv.fix(0.001)
    m.fs.valve_1.valve_opening.fix(1)
    m.fs.valve_2.valve_opening.fix(1)
    m.fs.tank.heat_duty.fix(0)
    m.fs.tank.control_volume.volume.fix(2.0)
    m.fs.ctrl.gain.fix(1e-6)
    m.fs.ctrl.time_i.fix(0.1)
    m.fs.ctrl.time_d.fix(0.1)
    m.fs.ctrl.setpoint.fix(3e5)
    # Initialize the model
    solver = pyo.SolverFactory("ipopt")
    solver.options = {'tol': 1e-6,
                      'linear_solver': "ma27",
                      'max_iter': 100}
    for t in m.fs.time:
        m.fs.valve_1.inlet.flow_mol = 100 # initial guess on flow
    # simple initialize
    m.fs.valve_1.initialize(outlvl=1)
    _set_port(m.fs.tank.inlet, m.fs.valve_1.outlet)
    m.fs.tank.initialize(outlvl=1)
    _set_port(m.fs.valve_2.inlet, m.fs.tank.outlet)
    m.fs.valve_2.initialize(outlvl=1)
    solver.solve(m, tee=True)
    # Return the model and solver
    return m, solver

@pytest.mark.slow
@pytest.mark.solver
@pytest.mark.skipif(not prop_available, reason="IAPWS not available")
@pytest.mark.skipif(not solver_available, reason="Solver not available")
def test_pid():
    """This test is pretty course-grained, but it should cover everything"""
    # Fisrt calculate the two steady states that should be achived in the test
    # don't worry these steady state problems solve super fast
    m_steady, solver = create_model()
    m_steady.fs.tank_pressure[0].fix(3e5)
    m_steady.fs.valve_1.valve_opening[0].unfix()
    solver.solve(m_steady, tee=True)
    s1_valve = pyo.value(m_steady.fs.valve_1.valve_opening[0])
    solver.solve(m_steady, tee=True)
    m_steady.fs.valve_1.inlet.pressure.fix(5.5e5)
    solver.solve(m_steady, tee=True)
    s2_valve = pyo.value(m_steady.fs.valve_1.valve_opening[0])
    # Next create a model for the 0 to 5 sec time period
    m_dynamic, solver = create_model(steady_state=False,
                                     time_set=[0,5], nfe=10, calc_integ=True)
    # Turn on control and solve since the setpoint is different than the
    # steady-state solution, stuff happens, also pressure step at t=5
    m_dynamic.fs.ctrl.activate()
    m_dynamic.fs.valve_1.valve_opening.unfix()
    m_dynamic.fs.valve_1.valve_opening[0].fix()
    # Add a step change right at the end of the interval, to make sure we can
    # get a continuous solution across the two models
    _add_inlet_pressure_step(m_dynamic, time=4.5, value=5.5e5)
    solver.solve(m_dynamic, tee=True)
    # Now create a model for the 5 to 10 second interval and set the inital
    # conditions of the first model to the final (unsteady) state of the
    # previous time interval model
    m_dynamic2, solver = create_model(steady_state=False,
                                     time_set=[5,10], nfe=10, calc_integ=False)

    m_dynamic2.fs.valve_1.valve_opening.fix(
        m_dynamic.fs.valve_1.valve_opening[5].value)
    m_dynamic2.fs.valve_1.inlet.enth_mol.fix(
        m_dynamic.fs.valve_1.inlet.enth_mol[5].value)
    m_dynamic2.fs.valve_1.inlet.pressure.fix(
        m_dynamic.fs.valve_1.inlet.pressure[5].value)
    for i in m_dynamic2.fs.tank.control_volume.material_accumulation:
        if i[0]==5:
            m_dynamic2.fs.tank.control_volume.material_accumulation[i].value =\
                m_dynamic.fs.tank.control_volume.material_accumulation[i].value
    for i in m_dynamic2.fs.tank.control_volume.enthalpy_accumulation:
        if i[0]==5:
            m_dynamic2.fs.tank.control_volume.enthalpy_accumulation[i].value =\
                m_dynamic.fs.tank.control_volume.enthalpy_accumulation[i].value
    m_dynamic2.fs.ctrl.err_d0.fix(pyo.value(m_dynamic.fs.ctrl.err_d[5]))
    m_dynamic2.fs.ctrl.err_i0.fix(pyo.value(m_dynamic.fs.ctrl.err_i[5]))
    # As a lazy form of initialization, solve the steady state problem before
    # turning on the controller.
    solver.solve(m_dynamic2, tee=True)
    # Now turn on control and solve again.
    m_dynamic2.fs.ctrl.activate()
    m_dynamic2.fs.valve_1.valve_opening.unfix()
    m_dynamic2.fs.valve_1.valve_opening[5].fix()
    solver.solve(m_dynamic2, tee=True)

    # Check that the model hit steady state about. The tolerance is loose
    # because I used a low nfe in an attempt to speed it up
    t = m_dynamic.fs.time.get_lower_element_boundary(4.45)
    assert s1_valve == pytest.approx(
        pyo.value(m_dynamic.fs.valve_1.valve_opening[t]), abs=5e-3)
    # The second should hit steady state pretty quick if the two models
    # line up smoothly.
    t = m_dynamic2.fs.time.get_lower_element_boundary(7)
    assert s2_valve == pytest.approx(
        pyo.value(m_dynamic2.fs.valve_1.valve_opening[t]), abs=5e-3)

    stitch_time = stitch_dynamic(m_dynamic.fs.time, m_dynamic2.fs.time)
    stitch_valve = stitch_dynamic(m_dynamic.fs.valve_1.valve_opening,
                                  m_dynamic2.fs.valve_1.valve_opening)

    # test the stitch function used to plot time dependent variables across
    # the two models
    for i, t in enumerate(m_dynamic.fs.time):
        assert t == stitch_time[i]
        assert m_dynamic.fs.valve_1.valve_opening[t] == stitch_valve[i]
    for j, t in enumerate(m_dynamic2.fs.time):
        i = j + len(m_dynamic.fs.time)
        assert t == stitch_time[i]
        assert m_dynamic2.fs.valve_1.valve_opening[t] == stitch_valve[i]



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module provides the base classes and methods for running convergence
evaluations on IDAES models. The convergence evaluation runs a given model over
a set of sample points to ensure reliable convergence over the parameter space.

The module requires the user to provide:
  - a set of inputs along with their lower bound, upper bound, mean,
and standard deviation.
  - an initialized Pyomo model
  - a Pyomo solver with appropriate options

The module executes convergence evaluation in two steps. In the first step, a
json file is created that containsa set of points sampled from the provided
inputs. This step only needs to be done once - up front. The second step, which
should be executed any time there is a major code change that could impact the
model, takes that set of sampled points and solves the model at each of the
points, collecting convergence statistics (success/failure, iterations, and
solution time).

This can be used as a tool to evaluate model convergence reliability over the
defined input space, or to verify that convergence performance is not
decreasing with framework and/or model changes.

In order to write a convergence evaluation for your model, you must inherit a
class from ConvergenceEvaluation, and implement three methods:

- get_specification: This method should create and return a
    ConvergenceEvaluationSpecification object. There are methods on
    ConvergenceEvaluationSpecification to add inputs. These inputs contain a
    string that identifies a Pyomo Param or Var object, the lower and upper
    bounds, and the mean and standard deviation to be used for sampling. When
    samples are generated, they are drawn from a normal distribution, and then
    truncated by the lower or upper bounds.
- get_initialized_model: This method should create and return a Pyomo model
    object that is already initialized and ready to be solved. This model will
    be modified according to the sampled inputs, and then it will be solved.
- get_solver: This method should return an instance of the Pyomo solver that
    will be used for the analysis.

There are methods to create the sample points file (on
ConvergenceEvaluationSpecification), to run a convergence evaluation
(run_convergence_evaluation), and print the results in table form
(print_convergence_statistics).

However, this package can also be executed using the command-line interface.
See the documentation in convergence.py for more information.
"""
# stdlib
from collections import OrderedDict
import getpass
import importlib as il
import json
import logging
import numpy as np
import sys
from io import StringIO
# pyomo
import pyutilib.services
from pyutilib.misc import capture_output
from pyomo.core import Param, Var
from pyomo.opt import TerminationCondition
from pyomo.common.log import LoggingIntercept
# idaes
import idaes.core.util.convergence.mpi_utils as mpiu
from idaes.dmf import resource


class ConvergenceEvaluationSpecification(object):
    def __init__(self):
        self._inputs = OrderedDict()

    def add_sampled_input(self, name, pyomo_path, lower, upper, mean, std):
        """
        Add an input that should be sampled when forming the set of
        specific points that need to be run for the convergence evaluation

        The input will be sampled assuming a normal distribution
        (with given mean and standard devation)
        truncated to the values given by lower and upper bounds

        Parameters
        ----------
        name : str
           The name of the input.
        pyomo_path : str
           A string representation of the path to the variable or parameter to
           be sampled. This string will be executed to retrieve the Pyomo
           component.
        lower : float
           A lower bound on the input variable or parameter.
        upper : float
           An upper bound on the input variable or parameter
        mean : float
           The mean value to use when generating samples
        std : float
           The standard deviation to use when generating samples

        Returns
        -------
           N/A
        """
        # ToDo: put some error checking here ... Maybe we should have the model
        # ToDo: already? Can use to check if the pyomo_path is valid? check if
        # ToDo: bounds will be violated?
        spec = OrderedDict()
        spec['pyomo_path'] = pyomo_path
        spec['lower'] = lower
        spec['upper'] = upper
        spec['mean'] = mean
        spec['std'] = std
        self._inputs[name] = spec

    @property
    def inputs(self):
        return self._inputs


class ConvergenceEvaluation(object):
    def __init__(self):
        self._sampling_specifications = list()

    def get_specification(self):
        """
        User should override this method to return an instance of the
        ConvergenceEvaluationSpecification for this particular model and test
        set.

        The basic flow for this method is:
           - Create a ConvergenceEvaluationSpecification
           - Call add_sampled_input for every input that should be varied.
           - return the ConvergenceEvaluationSpecification

        Returns
        -------
           ConvergenceEvaluationSpecification
        """
        raise NotImplementedError('Not implemented in the base class. This'
                                  ' should be overridden in the derived class')

    def get_initialized_model(self):
        """
        User should override this method to return an initialized model that is
        ready to solve. The convergence evaluation methods will change the
        values of parameters or variables according to the sampling
        specifications.

        Returns
        -------
           Pyomo model : return a Pyomo model object that is initialized and
                        ready to solve. This is the model object that will be
                        used in the evaluation.
        """
        raise NotImplementedError('Not implemented in the base class. This'
                                  ' should be overridden in the derived class')

    def get_solver(self):
        """
        User should create and return the solver that will be used for the
        convergence evaluation (including any necessary options)

        Returns
        -------
           Pyomo solver

        """
        # ToDo: We may want this to be standardized across all the models
        # within IDAES (should not need different solver options for different
        # unit models)
        raise NotImplementedError('Not implemented in the base class. This'
                                  ' should be overridden in the derived class')


def _class_import(class_path):
    # Note, this method assumes that everything in front
    # of the last dot is a module, followed by one class
    # I don't think this will work for nested classes
    tokens = class_path.split('.')
    modpath = '.'.join(tokens[0:-1])
    mod = il.import_module(modpath)
    ret_class = getattr(mod, tokens[-1])
    return ret_class


def _run_ipopt_with_stats(model, solver, max_iter=500, max_cpu_time=120):
    """
    Run the solver (must be ipopt) and return the convergence statistics

    Parameters
    ----------
    model : Pyomo model
       The pyomo model to be solved

    solver : Pyomo solver
       The pyomo solver to use - it must be ipopt, but with whichever options
       are preferred

    max_iter : int
       The maximum number of iterations to allow for ipopt

    max_cpu_time : int
       The maximum cpu time to allow for ipopt (in seconds)

    Returns
    -------
       Returns a tuple with (solve status object, bool (solve successful or
       not), number of iters, solve time)
    """
    # ToDo: Check that the "solver" is, in fact, IPOPT

    pyutilib.services.TempfileManager.push()
    tempfile = pyutilib.services.TempfileManager.create_tempfile(
                        suffix='ipopt_out',
                        text=True)
    opts = {'output_file': tempfile,
            'max_iter': max_iter,
            'max_cpu_time': max_cpu_time}

    status_obj = solver.solve(model, options=opts, tee=True)
    solved = True
    if status_obj.solver.termination_condition != TerminationCondition.optimal:
        solved = False

    iters = 0
    time = 0
    # parse the output file to get the iteration count, solver times, etc.
    with open(tempfile, 'r') as f:
        for line in f:
            if line.startswith('Number of Iterations....:'):
                tokens = line.split()
                iters = int(tokens[3])
            elif line.startswith(
                    'Total CPU secs in IPOPT (w/o function evaluations)   ='):
                tokens = line.split()
                time += float(tokens[9])
            elif line.startswith(
                    'Total CPU secs in NLP function evaluations           ='):
                tokens = line.split()
                time += float(tokens[8])

    pyutilib.services.TempfileManager.pop(remove=True)
    return status_obj, solved, iters, time


def _progress_bar(fraction, msg, length=20):
    length = length - 2
    n_complete = int(length*fraction)
    n_remaining = length - n_complete
    characters = ['*']*n_complete         # ['*' for i in range(n_complete)]
    characters.extend(['-']*n_remaining)  # ['-' for i in range(n_remaining)])
    sys.stdout.write('%5.1f%s [%s] %s\n' % (
                            fraction*100.0, '%', ''.join(characters), msg))
    sys.stdout.flush()


def _set_model_parameters_from_sample(model, inputs, sample_point):
    """
    This method takes the parameter values in sample_points and sets them on
    the model to prepare it for the new solve

    Parameters
    ----------
    model : Pyomo model
       The pyomo model that will be modified with values from the new
       sample_points

    inputs : dict
       The inputs dictionary from the sample-file

    sample_point : dict
       The dictionary of var and/or parameter values for the sample point

    Returns
    -------
       N/A
    """
    for k, v in sample_point.items():
        if k == '_name':
            # because parallel task manager does not handle dictionaries,
            # we add an _name entry to the sample point dictionary
            # ToDo: remove this when parallel task manager is fixed
            continue
        # k stores the "name" of the input
        # need to get the pyomo path of the input from the inputs structure
        pyomo_path = inputs[k]['pyomo_path']

        comp = model.find_component(pyomo_path)
        try:
            ctype = comp.type()
        except AttributeError:
            ctype = None

        if ctype is Param:
            if comp.is_constant():
                raise ValueError(
                        'Convergence testing found an input of type Param that'
                        'was not mutable. Please make sure all sampled inputs'
                        ' are either mutable params or fixed vars.')
            comp.set_value(v)
            # print('Just set', comp,'to', float(v))

        elif ctype is Var:
            if not comp.is_fixed():
                raise ValueError(
                        'Convergence testing found an input of type Var that'
                        'was not fixed. Please make sure all sampled inputs'
                        ' are either mutable params or fixed vars.')
            comp.set_value(float(v))
            # print('Just set', comp,'to', float(v))
        else:
            raise ValueError('Failed to find a valid input component (must be'
                             ' a fixed Var or a mutable Param). Instead,'
                             ' pyomo_path: {} returned: {}'
                             .format(pyomo_path, comp))


def write_sample_file(eval_spec,
                      filename,
                      convergence_evaluation_class_str,
                      n_points,
                      seed=None):
    """
    Samples the space of the inputs defined in the eval_spec, and creates a
    json file with all the points to be used in executing a convergence
    evaluation

    Parameters
    ----------
    filename : str
       The filename for the json file that will be created containing all the
       points to be run
    eval_spec : ConvergenceEvaluationSpecification
       The convergence evaluation specification object that we would like to
       sample
    convergence_evaluation_class_str : str
       Python string that identifies the convergence evaluation class for this
       specific evaluation. This is usually in the form of module.class_name.
    n_points : int
       The total number of points that should be created
    seed : int or None
       The seed to be used when generating samples. If set to None, then the
       seed is not set
    Returns
    -------
       N/A
    """
    if seed is not None:
        np.random.seed(seed)

    # build the samples
    samples = OrderedDict()
    for i in range(n_points):
        sample = samples['Sample-{}'.format(i+1)] = OrderedDict()
        for k, v in eval_spec.inputs.items():
            s = np.random.normal(loc=v['mean'], scale=v['std'])
            if s > v['upper']:
                s = v['upper']
            elif s < v['lower']:
                s = v['lower']
            sample[k] = s

    # create the dictionary storing all the necessary information
    jsondict = OrderedDict()
    jsondict['inputs'] = OrderedDict(eval_spec.inputs)
    jsondict['n_points'] = n_points
    jsondict['seed'] = seed
    jsondict['samples'] = samples
    jsondict['convergence_evaluation_class_str'] = \
        convergence_evaluation_class_str

    with open(filename, 'w') as fd:
        json.dump(jsondict, fd, indent=3)


def run_convergence_evaluation_from_sample_file(sample_file):
    # load the sample file
    try:
        with open(sample_file, 'r') as fd:
            jsondict = json.load(fd, object_pairs_hook=OrderedDict)
    except Exception as e:
        print('Error reading json file: {}'.format(str(e)))
        raise ValueError('Problem reading json file: {}'.format(sample_file))

    # create the convergence evaluation object
    convergence_evaluation_class_str = jsondict[
            'convergence_evaluation_class_str']
    try:
        convergence_evaluation_class = _class_import(
                                            convergence_evaluation_class_str)
        conv_eval = convergence_evaluation_class()
    except Exception as e:
        print('Error creating the class: {}. Error returned: {}'.format(
                    convergence_evaluation_class_str,
                    str(e)))
        raise ValueError(
                'Invalid value specified for convergence_evaluation_class_str:'
                '{} in sample file: {}'.format(
                        convergence_evaluation_class_str, sample_file))

    return run_convergence_evaluation(jsondict, conv_eval)


def run_convergence_evaluation(sample_file_dict, conv_eval):
    """
    Run convergence evaluation and generate the statistics based on information
    in the sample_file.

    Parameters
    ----------
    sample_file_dict : dict
        Dictionary created by ConvergenceEvaluationSpecification that contains
        the input and sample point information

    conv_eval : ConvergenceEvaluation
        The ConvergenceEvaluation object that should be used

    Returns
    -------
       N/A
    """
    inputs = sample_file_dict['inputs']
    samples = sample_file_dict['samples']

    # current parallel task manager code does not work with dictionaries, so
    # convert samples to a list
    # ToDo: fix and test parallel task manager with dictionaries and change
    # this
    samples_list = list()
    for k, v in samples.items():
        v['_name'] = k
        samples_list.append(v)
    n_samples = len(samples_list)

    task_mgr = mpiu.ParallelTaskManager(n_samples)
    local_samples_list = task_mgr.global_to_local_data(samples_list)

    results = list()
    for (si, ss) in enumerate(local_samples_list):
        sample_name = ss['_name']
        # print progress on the rank-0 process
        if task_mgr.is_root():
            _progress_bar(float(si) / float(len(local_samples_list)),
                          'Root Process: {}'.format(sample_name))

        # capture the output
        # ToDo: make this an option and turn off for single sample execution
        output_buffer = StringIO()
        with LoggingIntercept(output_buffer, 'idaes', logging.ERROR):
            with capture_output():  # as str_out:
                model = conv_eval.get_initialized_model()
                _set_model_parameters_from_sample(model, inputs, ss)
                solver = conv_eval.get_solver()
                (status_obj, solved, iters, time) = \
                    _run_ipopt_with_stats(model, solver)

        # run without output capture
        # model = conv_eval.get_initialized_model()
        # _set_model_parameters_from_sample(model, inputs, ss)
        # solver = conv_eval.get_solver()
        # (status_obj, solved, iters, time) = \
        #        _run_ipopt_with_stats(model, solver)

        if not solved:
            print('Sample: {} failed to converge.'.format(sample_name))

        results_dict = OrderedDict()
        results_dict['name'] = sample_name
        results_dict['sample_point'] = ss
        results_dict['solved'] = solved
        results_dict['iters'] = iters
        results_dict['time'] = time
        results.append(results_dict)

    global_results = task_mgr.gather_global_data(results)
    return inputs, samples, global_results


def save_convergence_statistics(inputs, results, dmf=None):
    s = Stats(results)
    if dmf is None:
        print_convergence_statistics(inputs, results, s)
    else:
        save_results_to_dmf(dmf, inputs, results, s)


class Stats(object):
    def __init__(self, results):
        self.notable_cases, self.failed_cases = [], []
        self.iters_successful = list()
        self.time_successful = list()

        # loop through and gather some data
        for r in results:
            if r['solved'] is True:
                self.iters_successful.append(r['iters'])
                self.time_successful.append(r['time'])
            else:
                self.failed_cases.append(r)
        # data for summary table
        self.iters_min = 0
        self.iters_mean = 0
        self.iters_std = 0
        self.iters_max = 0
        if len(self.iters_successful) > 0:
            self.iters_min = int(np.min(self.iters_successful))
            self.iters_mean = int(np.mean(self.iters_successful))
            self.iters_std = int(np.std(self.iters_successful))
            self.iters_max = int(np.max(self.iters_successful))
        self.time_min = 0
        self.time_mean = 0
        self.time_std = 0
        self.time_max = 0
        if len(self.time_successful) > 0:
            self.time_min = float(np.min(self.time_successful))
            self.time_mean = float(np.mean(self.time_successful))
            self.time_std = float(np.std(self.time_successful))
            self.time_max = float(np.max(self.time_successful))


def print_convergence_statistics(inputs, results, s):
    """
    Print the statistics returned from run_convergence_evaluation in a set of
    tables

    Parameters
    ----------
    inputs : dict
       The inputs dictionary returned by run_convergence_evaluation
    results : dict
       The results dictionary returned by run_convergence_evaluation

    Returns
    -------
       N/A
    """

    print()
    print('==== Scenario Statistics ====')
    print('%20s %10s %10s %10s %10s' % ('Parameter', 'Min', 'Mean',
                                        'StdDev', 'Max'))
    print('-'*64)
    for k, v in inputs.items():
        values = [results[i]['sample_point'][k] for i in range(len(results))]
        print('%20s %10g %10g %10g %10g' % (k, np.min(values),
                                            float(np.mean(values)),
                                            float(np.std(values)),
                                            float(np.max(values))))
    print('-'*64)

    print()
    print('==== Summary ====')
    print('Number of Successful Cases (solved=True): %.0f/%.0f' % (
            len(results) - len(s.failed_cases), len(results)))
    print('... Iterations      (min, -1std, mean, +1std, max):'
          '%5.0f, %5f, %5f, %5f, %5.0f'
          % (s.iters_min, s.iters_mean - s.iters_std,
             s.iters_mean, s.iters_mean + s.iters_std,
             s.iters_max))

    print('... Solver Time (s) (min, -1std, mean, +1std, max):'
          '%5f, %5f, %5f, %5f, %5f'
          % (s.time_min, s.time_mean - s.time_std,
             s.time_mean, s.time_mean + s.time_std,
             s.time_max))

    # print the detailed table
    print()
    print('==== Table of Results ====')
    print()
    print('%4s %20s %10s %10s %10s' % ('Flag', 'Name', 'Solved',
                                       'Iters', 'Time'))
    print('-' * 58)
    for r in results:
        flag = ''
        if r['solved'] is not True:
            flag = 'F'
        else:
            if r['time'] > s.time_mean + 2.0 * s.time_std and \
                    r['time'] > s.time_mean + 5.0:
                # add a more absolute check when s.time_std is small
                flag += 'T'
            if r['iters'] > s.iters_mean + 2.0 * s.iters_std and \
                    r['iters'] > s.iters_mean + 5:
                # add a more absolute check when s.iters_std is small
                flag += 'I'

        if flag != '':
            r['flag'] = flag
            s.notable_cases.append(r)

        print('%4s %20s %10s %10.0f %10.2f' % (
                flag, r['name'], r['solved'], r['iters'], r['time']))
        if r['solved'] is not True:
            s.failed_cases.append(r)
        else:
            s.iters_successful.append(r['iters'])
            s.time_successful.append(r['time'])

    print('-' * 58)

    print()
    print('==== Notable Cases ====')
    if len(s.notable_cases) == 0:
        print('... None')
    for c in s.notable_cases:
        msg = ''
        if c['flag'] == 'F':
            msg = 'failed solve'
        elif c['flag'] == 'T':
            msg = 'long runtime'
        elif c['flag'] == 'I':
            msg = 'high iteration count'
        elif c['flag'] == 'TI':
            msg = 'high iteration count; long runtime'

        print(c['name'], ':', msg)
        for k, v in c['sample_point'].items():
            if k == '_name':
                # we need this because the parallel task manager does not
                # handle dicts  and we add _name to identify the sample name
                # ToDo: fix this when the parallel task manager is extended to
                # handle dicts
                continue
            mean = inputs[k]['mean']
            std = inputs[k]['std']
            alpha = (v - mean) / std
            print('       %20s: %10g (%5.2f standard deviations'
                  ' above/below the mean)' % (k, v, alpha))


def save_results_to_dmf(dmf, inputs, results, stats):
    """Save results of run, along with stats, to DMF.

    Args:
        dmf (DMF): Data management framework object
        inputs (dict): Run inputs
        results (dict): Run results
        stats (Stats): Calculated result statistics

    Returns:
        None
    """
    d = {}  # data
    x = {}
    for k, v in inputs.items():
        values = [results[i]['sample_point'][k] for i in range(len(results))]
        x[k] = {
            'min': float(np.min(values)),
            'mean': float(np.mean(values)),
            'stdev': float(np.std(values)),
            'max': float(np.max(values))
        }
    d['scenarios'] = x
    d['summary'] = {
        'cases_success': len(results) - len(stats.failed_cases),
        'cases_total': len(results),
        'iters': {
            'min': stats.iters_min,
            '-1std': stats.iters_mean - stats.iters_std,
            'mean': stats.iters_mean,
            '+1std': stats.iters_mean + stats.iters_std,
            'max': stats.iters_max
        },
        'time': {
            'min': stats.time_min,
            '-1std': stats.time_mean - stats.time_std,
            'mean': stats.time_mean,
            '+1std': stats.time_mean + stats.time_std,
            'max': stats.time_max
        }
    }
    tbl = []
    for r in results:
        notable = False
        solved = r['solved']
        item = {
            'solved': solved,
            'name': r['name'],
            'iters': r['iters'],
            'time': r['time']
        }
        for vtype in 'time', 'iters':
            mean, std = [getattr(stats, a) for a in ('{}_mean'.format(vtype),
                                                     '{}_std'.format(vtype))]
            if r[vtype] > mean + 2. * std and r[vtype] > mean + 5:
                item['outlier_{}'.format(vtype)] = True
                notable = True
        if not solved:
            stats.failed_cases.append(r)
            notable = True
        else:
            stats.iters_successful.append(r['iters'])
            stats.time_successful.append(r['time'])
        tbl.append(item)
        if notable:
            stats.notable_cases.append(r)
    d['results_table'] = tbl
    # notable cases
    cases = []
    for r in stats.notable_cases:
        msg, is_outlier = [], False
        if not r['solved']:
            msg.append('failed solve')
        else:
            is_outlier = True
            if r['outlier_time']:
                msg.append('long runtime')
            if r['outlier_iters']:
                msg.append('high iteration count')
        case = {'messages': msg}
        if is_outlier:
            outliers = {}
            for k, v in r['sample_point'].items():
                if k == '_name':
                    # we need this because the parallel task manager does not
                    # handle dicts and we add _name to identify the sample name
                    # TODO: Fix this when the parallel task manager is extended
                    # TODO: to handle dicts
                    continue
                mean = inputs[k]['mean']
                std = inputs[k]['std']
                alpha = (v - mean) / std
                outliers[k] = (v, alpha)
            case['outliers'] = outliers
        else:
            case['outliers'] = []
        cases.append(case)
    d['notable_cases'] = cases
    # Build and save resource object
    rsrc = resource.Resource(value={
        'name': 'convergence_results',
        'desc': 'statistics returned from run_convergence_evaluation',
        'creator': {'name': getpass.getuser()},
        'data': d}, type_=resource.TY_DATA)
    dmf.add(rsrc)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.vis.plotbase
"""
# third-party
import pytest
import pandas as pd
# package
from idaes.vis import plotbase


def test_plotbase_construct_abstract():
    assert pytest.raises(TypeError, plotbase.PlotBase, 'foo')


def test_plotregistry_construct():
    reg = plotbase.PlotRegistry()


def test_plotregistry_register():
    reg = plotbase.PlotRegistry()
    reg.remove_all()
    obj = "hello"
    plot_class = plotbase.PlotBase
    reg.register(obj, "test1", plot_class)
    reg.remove_all()


def setup_plot(plot, obj):
    return plot, obj


def test_plotregistry_register_setup():
    reg = plotbase.PlotRegistry()
    reg.remove_all()
    obj = "hello"
    plot_class = plotbase.PlotBase
    reg.register(obj, "test1", plot_class, setup_fn=setup_plot)


def test_plotregistry_register_overwrite():
    reg = plotbase.PlotRegistry()
    reg.remove_all()
    obj = "hello"
    plot_class = plotbase.PlotBase
    reg.register(obj, "test1", plot_class)
    # without overwrite: KeyError
    assert pytest.raises(
        KeyError, reg.register, obj, "test1", plot_class, dict(setup_fn=setup_plot)
    )
    # with overwrite: ok
    reg.register(obj, "test1", plot_class, overwrite=True)


def test_plotregistry_get():
    reg = plotbase.PlotRegistry()
    reg.remove_all()
    obj = "hello"
    plot_class = plotbase.PlotBase
    reg.register(obj, "test1", plot_class, setup_fn=setup_plot)
    result = reg.get(obj, "test1")
    assert result == (plot_class, obj)


class DummyPlot(plotbase.PlotBase):
    def __init__(self):
        super().__init__(None)

    def annotate(self, x, y, label: str):
        return

    def resize(self, height: int = -1, width: int = -1):
        return

    def save(self, destination: str):
        return 'filename'

    def show(self, in_notebook=True):
        return


@pytest.fixture
def plotdf():
    return pd.DataFrame({'x': [1, 2], 'y1': [1, 2], 'y2': [1, 2]})


def test_validate(plotdf):
    # reject
    for bad_args, bad_kwargs in [
        ((None, None, None), {}),
        ((None, None, None), {'legend': 'i_am'}),
        ((None, 'x', []), {}),
        ((plotdf, 'x', ['z']), {}),
        ((plotdf, 'foo', ['y1']), {}),
    ]:
        plot = DummyPlot()
        result, msg = plot.validate(*bad_args, **bad_kwargs)
        assert result is False
    # pass
    for good_args, good_kwargs in [
        ((plotdf, 'x', ['y1']), {}),
        ((plotdf, 'x', ['y1', 'y2']), {}),
        ((plotdf, 'x', ['y1']), {'legend': 'foo'}),
    ]:
        plot = DummyPlot()
        result, msg = plot.validate(*good_args, **good_kwargs)
        assert result is True



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
General purpose separator block for IDAES models
"""

from enum import Enum
from pandas import DataFrame

from pyomo.environ import (Constraint,
                           Expression,
                           Param,
                           Reals,
                           Reference,
                           Set,
                           SolverFactory,
                           TerminationCondition,
                           Var,
                           value)
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In

from idaes.core import (declare_process_block_class,
                        UnitModelBlockData,
                        useDefault,
                        MaterialBalanceType)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_state_block,
                                    list_of_strings)
from idaes.core.util.exceptions import (BurntToast,
                                        ConfigurationError,
                                        PropertyNotSupportedError)
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.misc import add_object_reference
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

__author__ = "Andrew Lee"


# Set up logger
_log = getIdaesLogger(__name__)


# Enumerate options for balances
class SplittingType(Enum):
    totalFlow = 1
    phaseFlow = 2
    componentFlow = 3
    phaseComponentFlow = 4


class EnergySplittingType(Enum):
    equal_temperature = 1
    equal_molar_enthalpy = 2
    enthalpy_split = 3


@declare_process_block_class("Separator")
class SeparatorData(UnitModelBlockData):
    """
    This is a general purpose model for a Separator block with the IDAES
    modeling framework. This block can be used either as a stand-alone
    Separator unit operation, or as a sub-model within another unit operation.

    This model creates a number of StateBlocks to represent the outgoing
    streams, then writes a set of phase-component material balances, an
    overall enthalpy balance (2 options), and a momentum balance (2 options)
    linked to a mixed-state StateBlock. The mixed-state StateBlock can either
    be specified by the user (allowing use as a sub-model), or created by the
    Separator.

    When being used as a sub-model, Separator should only be used when a
    set of new StateBlocks are required for the streams to be separated. It
    should not be used to separate streams to go to mutiple ControlVolumes in a
    single unit model - in these cases the unit model developer should write
    their own splitting equations.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
**default** = False. Product blocks are always steady-state."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Product blocks do not contain holdup, thus this must be
False."""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for mixer",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("outlet_list", ConfigValue(
        domain=list_of_strings,
        description="List of outlet names",
        doc="""A list containing names of outlets,
**default** - None.
**Valid values:** {
**None** - use num_outlets argument,
**list** - a list of names to use for outlets.}"""))
    CONFIG.declare("num_outlets", ConfigValue(
        domain=int,
        description="Number of outlets to unit",
        doc="""Argument indicating number (int) of outlets to construct, not
used if outlet_list arg is provided,
**default** - None.
**Valid values:** {
**None** - use outlet_list arg instead, or default to 2 if neither argument
provided,
**int** - number of outlets to create (will be named with sequential integers
from 1 to num_outlets).}"""))
    CONFIG.declare("split_basis", ConfigValue(
        default=SplittingType.totalFlow,
        domain=SplittingType,
        description="Basis for splitting stream",
        doc="""Argument indicating basis to use for splitting mixed stream,
**default** - SplittingType.totalFlow.
**Valid values:** {
**SplittingType.totalFlow** - split based on total flow (split
fraction indexed only by time and outlet),
**SplittingType.phaseFlow** - split based on phase flows (split fraction
indexed by time, outlet and phase),
**SplittingType.componentFlow** - split based on component flows (split
fraction indexed by time, outlet and components),
**SplittingType.phaseComponentFlow** - split based on phase-component flows (
split fraction indexed by both time, outlet, phase and components).}"""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Calculate phase equilibrium in mixed stream",
        doc="""Argument indicating whether phase equilibrium should be
calculated for the resulting mixed stream,
**default** - False.
**Valid values:** {
**True** - calculate phase equilibrium in mixed stream,
**False** - do not calculate equilibrium in mixed stream.}"""))
    CONFIG.declare("energy_split_basis", ConfigValue(
        default=EnergySplittingType.equal_temperature,
        domain=EnergySplittingType,
        description="Type of constraint to write for energy splitting",
        doc="""Argument indicating basis to use for splitting energy this is
not used for when ideal_separation == True.
**default** - EnergySplittingType.equal_temperature.
**Valid values:** {
**EnergySplittingType.equal_temperature** - outlet temperatures equal inlet
**EnergySplittingType.equal_molar_enthalpy** - oulet molar enthalpies equal
inlet,
**EnergySplittingType.enthalpy_split** - apply split fractions to enthalpy
flows. Does not work with component or phase-component splitting.}"""))
    CONFIG.declare("ideal_separation", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Ideal splitting flag",
        doc="""Argument indicating whether ideal splitting should be used.
Ideal splitting assumes perfect spearation of material, and attempts to
avoid duplication of StateBlocks by directly partitioning outlet flows to
ports,
**default** - False.
**Valid values:** {
**True** - use ideal splitting methods. Cannot be combined with
has_phase_equilibrium = True,
**False** - use explicit splitting equations with split fractions.}"""))
    CONFIG.declare("ideal_split_map", ConfigValue(
        domain=dict,
        description="Ideal splitting partitioning map",
        doc="""Dictionary containing information on how extensive variables
should be partitioned when using ideal splitting (ideal_separation = True).
**default** - None.
**Valid values:** {
**dict** with keys of indexing set members and values indicating which outlet
this combination of keys should be partitioned to.
E.g. {("Vap", "H2"): "outlet_1"}}"""))
    CONFIG.declare("mixed_state_block", ConfigValue(
        domain=is_state_block,
        description="Existing StateBlock to use as mixed stream",
        doc="""An existing state block to use as the source stream from the
Separator block,
**default** - None.
**Valid values:** {
**None** - create a new StateBlock for the mixed stream,
**StateBlock** - a StateBock to use as the source for the mixed stream.}"""))
    CONFIG.declare("construct_ports", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Construct inlet and outlet Port objects",
        doc="""Argument indicating whether model should construct Port objects
linked the mixed state and all outlet states,
**default** - True.
**Valid values:** {
**True** - construct Ports for all states,
**False** - do not construct Ports."""))

    def build(self):
        """
        General build method for SeparatorData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        # Call super.build()
        super(SeparatorData, self).build()

        self._validate_config_arguments()

        # Call setup methods from ControlVolumeBlockData
        self._get_property_package()
        self._get_indexing_sets()

        # Create list of inlet names
        outlet_list = self.create_outlet_list()

        if self.config.mixed_state_block is None:
            mixed_block = self.add_mixed_state_block()
        else:
            mixed_block = self.get_mixed_state_block()

        # Add inlet port
        self.add_inlet_port_objects(mixed_block)

        # Construct splitter based on ideal_separation argument
        if self.config.ideal_separation:
            # Use ideal partitioning method
            self.partition_outlet_flows(mixed_block, outlet_list)
        else:
            # Otherwise, Build StateBlocks for outlet
            outlet_blocks = self.add_outlet_state_blocks(outlet_list)

            # Add split fractions
            self.add_split_fractions(outlet_list)

            # Construct splitting equations
            self.add_material_splitting_constraints(mixed_block)
            self.add_energy_splitting_constraints(mixed_block)
            self.add_momentum_splitting_constraints(mixed_block)

            # Construct outlet port objects
            self.add_outlet_port_objects(outlet_list, outlet_blocks)

    def _validate_config_arguments(self):
        if self.config.has_phase_equilibrium and self.config.ideal_separation:
            raise ConfigurationError(
                    """{} recieved arguments has_phase_equilibrium = True and
                    ideal_separation = True. These arguments are incompatible
                    with each other, and you should choose one or the other."""
                    .format(self.name))

    def create_outlet_list(self):
        """
        Create list of outlet stream names based on config arguments.

        Returns:
            list of strings
        """
        if (self.config.outlet_list is not None and
                self.config.num_outlets is not None):
            # If both arguments provided and not consistent, raise Exception
            if len(self.config.outlet_list) != self.config.num_outlets:
                raise ConfigurationError(
                        "{} Separator provided with both outlet_list and "
                        "num_outlets arguments, which were not consistent ("
                        "length of outlet_list was not equal to num_outlets). "
                        "Please check your arguments for consistency, and "
                        "note that it is only necessry to provide one of "
                        "these arguments.".format(self.name))
        elif (self.config.outlet_list is None and
              self.config.num_outlets is None):
            # If no arguments provided for outlets, default to num_outlets = 2
            self.config.num_outlets = 2

        # Create a list of names for outlet StateBlocks
        if self.config.outlet_list is not None:
            outlet_list = self.config.outlet_list
        else:
            outlet_list = ['outlet_' + str(n)
                           for n in range(1, self.config.num_outlets+1)]

        return outlet_list

    def add_outlet_state_blocks(self, outlet_list):
        """
        Construct StateBlocks for all outlet streams.

        Args:
            list of strings to use as StateBlock names

        Returns:
            list of StateBlocks
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False

        # Create empty list to hold StateBlocks for return
        outlet_blocks = []

        # Create an instance of StateBlock for all outlets
        for o in outlet_list:
            o_obj = self.config.property_package.state_block_class(
                        self.flowsheet().config.time,
                        doc="Material properties at outlet",
                        default=tmp_dict)

            setattr(self, o+"_state", o_obj)

            outlet_blocks.append(getattr(self, o+"_state"))

        return outlet_blocks

    def add_mixed_state_block(self):
        """
        Constructs StateBlock to represent mixed stream.

        Returns:
            New StateBlock object
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True

        self.mixed_state = self.config.property_package.state_block_class(
                                self.flowsheet().config.time,
                                doc="Material properties of mixed stream",
                                default=tmp_dict)

        return self.mixed_state

    def get_mixed_state_block(self):
        """
        Validates StateBlock provided in user arguments for mixed stream.

        Returns:
            The user-provided StateBlock or an Exception
        """
        # Sanity check to make sure method is not called when arg missing
        if self.config.mixed_state_block is None:
            raise BurntToast("{} get_mixed_state_block method called when "
                             "mixed_state_block argument is None. This should "
                             "not happen.".format(self.name))

        # Check that the user-provided StateBlock uses the same prop pack
        if (self.config.mixed_state_block[
                    self.flowsheet().config.time.first()].config.parameters
                != self.config.property_package):
            raise ConfigurationError(
                    "{} StateBlock provided in mixed_state_block argument "
                    " does not come from the same property package as "
                    "provided in the property_package argument. All "
                    "StateBlocks within a Separator must use the same "
                    "property package.".format(self.name))

        return self.config.mixed_state_block

    def add_inlet_port_objects(self, mixed_block):
        """
        Adds inlet Port object if required.

        Args:
            a mixed state StateBlock object

        Returns:
            None
        """
        if self.config.construct_ports is True:
            self.add_port(name="inlet", block=mixed_block, doc="Inlet Port")

    def add_outlet_port_objects(self, outlet_list, outlet_blocks):
        """
        Adds outlet Port objects if required.

        Args:
            a list of outlet StateBlock objects

        Returns:
            None
        """
        if self.config.construct_ports is True:
            # Add ports
            for p in outlet_list:
                o_state = getattr(self, p+"_state")
                self.add_port(name=p, block=o_state, doc="Outlet Port")

    def add_split_fractions(self, outlet_list):
        """
        Creates outlet Port objects and tries to partiton mixed stream flows
        between these

        Args:
            StateBlock representing the mixed flow to be split
            a list of names for outlets

        Returns:
            None
        """
        self.outlet_idx = Set(initialize=outlet_list)

        if self.config.split_basis == SplittingType.totalFlow:
            sf_idx = [self.flowsheet().config.time, self.outlet_idx]
            sf_sum_idx = [self.flowsheet().config.time]
        elif self.config.split_basis == SplittingType.phaseFlow:
            sf_idx = [self.flowsheet().config.time,
                      self.outlet_idx, self.config.property_package.phase_list]
            sf_sum_idx = [self.flowsheet().config.time,
                          self.config.property_package.phase_list]
        elif self.config.split_basis == SplittingType.componentFlow:
            sf_idx = [self.flowsheet().config.time, self.outlet_idx,
                      self.config.property_package.component_list]
            sf_sum_idx = [self.flowsheet().config.time,
                          self.config.property_package.component_list]
        elif self.config.split_basis == SplittingType.phaseComponentFlow:
            sf_idx = [self.flowsheet().config.time,
                      self.outlet_idx,
                      self.config.property_package.phase_list,
                      self.config.property_package.component_list]
            sf_sum_idx = [self.flowsheet().config.time,
                          self.config.property_package.phase_list,
                          self.config.property_package.component_list]
        else:
            raise BurntToast("{} split_basis has unexpected value. This "
                             "should not happen.".format(self.name))

        # Create split fraction variable
        self.split_fraction = Var(*sf_idx,
                                  initialize=0.5,
                                  doc="Outlet split fractions")

        # Add constraint that split fractions sum to 1
        def sum_sf_rule(b, t, *args):
            return 1 == sum(b.split_fraction[t, o, args]
                            for o in self.outlet_idx)
        self.sum_split_frac = Constraint(*sf_sum_idx, rule=sum_sf_rule)

    def add_material_splitting_constraints(self, mixed_block):
        """
        Creates constraints for splitting the material flows
        """
        def sf(t, o, p, j):
            if self.config.split_basis == SplittingType.totalFlow:
                return self.split_fraction[t, o]
            elif self.config.split_basis == SplittingType.phaseFlow:
                return self.split_fraction[t, o, p]
            elif self.config.split_basis == SplittingType.componentFlow:
                return self.split_fraction[t, o, j]
            elif self.config.split_basis == SplittingType.phaseComponentFlow:
                return self.split_fraction[t, o, p, j]

        mb_type = self.config.material_balance_type
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = \
                mixed_block[t_ref].default_material_balance_type()

        if mb_type == MaterialBalanceType.componentPhase:
            if self.config.has_phase_equilibrium is True:
                # Get units from property package
                units = {}
                for u in ['holdup', 'time']:
                    try:
                        units[u] = (self.config.property_package
                                    .get_metadata().default_units[u])
                    except KeyError:
                        units[u] = '-'

                try:
                    self.phase_equilibrium_generation = Var(
                            self.flowsheet().config.time,
                            self.outlet_idx,
                            self.config.property_package.phase_equilibrium_idx,
                            domain=Reals,
                            doc="Amount of generation in unit by phase "
                                "equilibria [{}/{}]"
                                .format(units['holdup'], units['time']))
                except AttributeError:
                    raise PropertyNotSupportedError(
                        "{} Property package does not contain a list of phase "
                        "equilibrium reactions (phase_equilibrium_idx), thus "
                        "does not support phase equilibrium."
                        .format(self.name))

            # Define terms to use in mixing equation
            def phase_equilibrium_term(b, t, o, p, j):
                if self.config.has_phase_equilibrium:
                    sd = {}
                    sblock = mixed_block[t]
                    for r in b.config.property_package.phase_equilibrium_idx:
                        if sblock._params.phase_equilibrium_list[r][0] == j:
                            if sblock._params \
                                    .phase_equilibrium_list[r][1][0] == p:
                                sd[r] = 1
                            elif sblock._params \
                                    .phase_equilibrium_list[r][1][1] == p:
                                sd[r] = -1
                            else:
                                sd[r] = 0
                        else:
                            sd[r] = 0

                    return sum(b.phase_equilibrium_generation[t, o, r]*sd[r]
                               for r in
                               b.config.property_package.phase_equilibrium_idx)
                else:
                    return 0

            @self.Constraint(self.flowsheet().config.time,
                             self.outlet_idx,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Material splitting equations")
            def material_splitting_eqn(b, t, o, p, j):
                o_block = getattr(self, o+"_state")
                return (sf(t, o, p, j) *
                        mixed_block[t].get_material_flow_terms(p, j) ==
                        o_block[t].get_material_flow_terms(p, j) -
                        phase_equilibrium_term(b, t, o, p, j))
        elif mb_type == MaterialBalanceType.componentTotal:
            @self.Constraint(self.flowsheet().config.time,
                             self.outlet_idx,
                             self.config.property_package.component_list,
                             doc="Material splitting equations")
            def material_splitting_eqn(b, t, o, j):
                o_block = getattr(self, o+"_state")
                return (sum(sf(t, o, p, j) *
                            mixed_block[t].get_material_flow_terms(p, j) for p
                            in b.config.property_package.phase_list) ==
                        sum(o_block[t].get_material_flow_terms(p, j) for p in
                            b.config.property_package.phase_list))
        elif mb_type == MaterialBalanceType.total:
            @self.Constraint(self.flowsheet().config.time,
                             self.outlet_idx,
                             doc="Material splitting equations")
            def material_splitting_eqn(b, t, o):
                o_block = getattr(self, o+"_state")
                return (
                    sum(sum(sf(t, o, p, j) *
                            mixed_block[t].get_material_flow_terms(p, j) for j
                            in b.config.property_package.component_list)
                        for p in b.config.property_package.phase_list) ==
                    sum(sum(o_block[t].get_material_flow_terms(p, j) for j
                            in b.config.property_package.component_list)
                        for p in b.config.property_package.phase_list))
        elif mb_type == MaterialBalanceType.elementTotal:
            raise ConfigurationError("{} Separators do not support elemental "
                                     "material balances.".format(self.name))
        elif mb_type == MaterialBalanceType.none:
            pass
        else:
            raise BurntToast("{} Separator received unrecognised value for "
                             "material_balance_type. This should not happen, "
                             "please report this bug to the IDAES developers."
                             .format(self.name))

    def add_energy_splitting_constraints(self, mixed_block):
        """
        Creates constraints for splitting the energy flows - done by equating
        temperatures in outlets.
        """
        if self.config.energy_split_basis == \
                EnergySplittingType.equal_temperature:
            @self.Constraint(self.flowsheet().config.time,
                             self.outlet_idx,
                             doc="Temperature equality constraint")
            def temperature_equality_eqn(b, t, o):
                o_block = getattr(self, o+"_state")
                return mixed_block[t].temperature == o_block[t].temperature
        elif self.config.energy_split_basis == \
                EnergySplittingType.equal_molar_enthalpy:
            @self.Constraint(self.flowsheet().config.time,
                             self.outlet_idx,
                             doc="Molar enthalpy equality constraint")
            def molar_enthalpy_equality_eqn(b, t, o):
                o_block = getattr(self, o+"_state")
                return mixed_block[t].enth_mol == o_block[t].enth_mol
        elif self.config.energy_split_basis == \
                EnergySplittingType.enthalpy_split:
            # Validate split fraction type
            if (self.config.split_basis == SplittingType.phaseComponentFlow or
                    self.config.split_basis == SplittingType.componentFlow):
                raise ConfigurationError(
                        "{} Cannot use energy_split_basis == enthalpy_split "
                        "with split_basis == component or phaseComponent."
                        .format(self.name))

            def sf(t, o, p):
                if self.config.split_basis == SplittingType.totalFlow:
                    return self.split_fraction[t, o]
                elif self.config.split_basis == SplittingType.phaseFlow:
                    return self.split_fraction[t, o, p]

            @self.Constraint(self.flowsheet().config.time,
                             self.outlet_idx,
                             doc="Molar enthalpy splitting constraint")
            def molar_enthalpy_splitting_eqn(b, t, o):
                o_block = getattr(self, o+"_state")
                return (sum(mixed_block[t].get_enthalpy_flow_terms(p) *
                            sf(t, o, p)
                            for p in b.config.property_package.phase_list) ==
                        sum(o_block[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list))
        else:
            raise BurntToast("{} received unrecognised value for "
                             "energy_split_basis. This should never happen, so"
                             " please contact the IDAES developers with this "
                             "bug.".format(self.name))

    def add_momentum_splitting_constraints(self, mixed_block):
        """
        Creates constraints for splitting the momentum flows - done by equating
        pressures in outlets.
        """
        @self.Constraint(self.flowsheet().config.time,
                         self.outlet_idx,
                         doc="Pressure equality constraint")
        def pressure_equality_eqn(b, t, o):
            o_block = getattr(self, o+"_state")
            return mixed_block[t].pressure == o_block[t].pressure

    def partition_outlet_flows(self, mb, outlet_list):
        """
        Creates outlet Port objects and tries to partiton mixed stream flows
        between these

        Args:
            StateBlock representing the mixed flow to be split
            a list of names for outlets

        Returns:
            None
        """
        # Check arguments
        if self.config.construct_ports is False:
            raise ConfigurationError("{} cannot have and ideal separator "
                                     "(ideal_separation = True) with "
                                     "construct_ports = False."
                                     .format(self.name))
        if self.config.split_basis == SplittingType.totalFlow:
            raise ConfigurationError("{} cannot do an ideal separation based "
                                     "on total flow.".format(self.name))
        if self.config.ideal_split_map is None:
            raise ConfigurationError("{} was not provided with an "
                                     "ideal_split_map argument which is "
                                     "necessary for doing an ideal_separation."
                                     .format(self.name))

        # Validate split map
        split_map = self.config.ideal_split_map
        idx_list = []
        if self.config.split_basis == SplittingType.phaseFlow:
            for p in self.config.property_package.phase_list:
                idx_list.append((p))

            if len(idx_list) != len(split_map):
                raise ConfigurationError(
                        "{} ideal_split_map does not match with "
                        "split_basis chosen. ideal_split_map must"
                        " have a key for each combination of indices."
                        .format(self.name))
            for k in idx_list:
                if k not in split_map:
                    raise ConfigurationError(
                        "{} ideal_split_map does not match with "
                        "split_basis chosen. ideal_split_map must"
                        " have a key for each combination of indices."
                        .format(self.name))

        elif self.config.split_basis == SplittingType.componentFlow:
            for j in self.config.property_package.component_list:
                idx_list.append((j))

            if len(idx_list) != len(split_map):
                raise ConfigurationError(
                        "{} ideal_split_map does not match with "
                        "split_basis chosen. ideal_split_map must"
                        " have a key for each component."
                        .format(self.name))
        elif self.config.split_basis == SplittingType.phaseComponentFlow:
            for p in self.config.property_package.phase_list:
                for j in self.config.property_package.component_list:
                    idx_list.append((p, j))

            if len(idx_list) != len(split_map):
                raise ConfigurationError(
                        "{} ideal_split_map does not match with "
                        "split_basis chosen. ideal_split_map must"
                        " have a key for each phase-component pair."
                        .format(self.name))

        # Check that no. outlets matches split_basis
        if len(outlet_list) != len(idx_list):
            raise ConfigurationError(
                        "{} Cannot perform ideal separation. Must have one "
                        "outlet for each possible combination of the "
                        "chosen split_basis."
                        .format(self.name))

        # Create tolerance Parameter for 0 flow outlets
        self.eps = Param(default=1e-8, mutable=True)

        # Get list of port members
        s_vars = mb[self.flowsheet().config.time.first()].define_port_members()

        # Add empty Port objects
        for o in outlet_list:
            p_obj = Port(noruleinit=True,
                         doc="Outlet Port")
            setattr(self, o, p_obj)

            # Iterate over members to create References or Expressions
            for s in s_vars:
                # Get local variable name of component
                l_name = s_vars[s].local_name

                if l_name == "pressure" or l_name == "temperature":
                    # Assume outlets same as mixed flow - make Reference
                    e_obj = Reference(mb[:].component(l_name))

                elif (l_name.startswith("mole_frac") or
                      l_name.startswith("mass_frac")):
                    # Mole and mass frac need special handling
                    if "_phase" in l_name:
                        def e_rule(b, t, p, j):
                            if self.config.split_basis == \
                                        SplittingType.phaseFlow:
                                return s_vars[s][p, j]
                            elif self.config.split_basis == \
                                    SplittingType.componentFlow:
                                if split_map[j] == o:
                                    return 1
                                else:
                                    return self.eps
                            elif self.config.split_basis == \
                                    SplittingType.phaseComponentFlow:
                                for ps in self.config.property_package \
                                        .phase_list:
                                    if split_map[ps, j] == o:
                                        return 1
                                return self.eps
                            else:
                                raise BurntToast(
                                        "{} This should not happen. Please "
                                        "report this bug to the IDAES "
                                        "developers.".format(self.name))

                        e_obj = Expression(
                                self.flowsheet().config.time,
                                self.config.property_package.phase_list,
                                self.config.property_package.component_list,
                                rule=e_rule)

                    else:
                        if self.config.split_basis == \
                                    SplittingType.componentFlow:
                            def e_rule(b, t, j):
                                if split_map[j] == o:
                                    return 1
                                # else:
                                return self.eps

                        elif self.config.split_basis == \
                                SplittingType.phaseComponentFlow:
                            def e_rule(b, t, j):
                                if any(split_map[p, j] == o for p in
                                       self.config.property_package
                                       .phase_list):
                                    return 1
                                # else:
                                return self.eps

                        else:
                            def e_rule(b, t, j):
                                mfp = mb[t].component(
                                        l_name.replace("_comp", "_phase_comp"))

                                if mfp is None:
                                    raise AttributeError(
                                        "{} Cannot use ideal splitting with "
                                        "this property package. Package uses "
                                        "indexed port member {} which does not"
                                        " have the correct indexing sets, and "
                                        "an equivalent variable with correct "
                                        "indexing sets is not available."
                                        .format(self.name, s))

                                for p in self.config.property_package \
                                        .phase_list:
                                    if self.config.split_basis == \
                                            SplittingType.phaseFlow:
                                        s_check = split_map[p]
                                    else:
                                        raise BurntToast(
                                            "{} This should not happen. Please"
                                            " report this bug to the IDAES "
                                            "developers.".format(self.name))

                                    if s_check == o:
                                        return mfp[p, j]
                                # else:
                                return self.eps

                        e_obj = Expression(
                                self.flowsheet().config.time,
                                self.config.property_package.component_list,
                                rule=e_rule)

                elif l_name.endswith("_phase_comp"):
                    def e_rule(b, t, p, j):
                        if self.config.split_basis == \
                                SplittingType.phaseFlow:
                            s_check = split_map[p]
                        elif self.config.split_basis == \
                                SplittingType.componentFlow:
                            s_check = split_map[j]
                        elif self.config.split_basis == \
                                SplittingType.phaseComponentFlow:
                            s_check = split_map[p, j]
                        else:
                            raise BurntToast(
                                    "{} This should not happen. Please"
                                    " report this bug to the IDAES "
                                    "developers.".format(self.name))

                        if s_check == o:
                            return mb[t].component(l_name)[p, j]
                        else:
                            return self.eps

                    e_obj = Expression(
                                self.flowsheet().config.time,
                                self.config.property_package.phase_list,
                                self.config.property_package.component_list,
                                rule=e_rule)

                elif l_name.endswith("_phase"):
                    if self.config.split_basis == \
                                    SplittingType.phaseFlow:
                        def e_rule(b, t, p):
                            if split_map[p] == o:
                                return mb[t].component(l_name)[p]
                            else:
                                return self.eps

                    else:
                        def e_rule(b, t, p):
                            mfp = mb[t].component(l_name+"_comp")

                            if mfp is None:
                                raise AttributeError(
                                    "{} Cannot use ideal splitting with this "
                                    "property package. Package uses indexed "
                                    "port member {} which does not have the "
                                    "correct indexing sets, and an equivalent "
                                    "variable with correct indexing sets is "
                                    "not available."
                                    .format(self.name, s))

                            for j in self.config.property_package \
                                    .component_list:
                                if self.config.split_basis == \
                                        SplittingType.componentFlow:
                                    s_check = split_map[j]
                                elif self.config.split_basis == \
                                        SplittingType.phaseComponentFlow:
                                    s_check = split_map[p, j]
                                else:
                                    raise BurntToast(
                                        "{} This should not happen. Please"
                                        " report this bug to the IDAES "
                                        "developers.".format(self.name))

                                if s_check == o:
                                    return mfp[p, j]
                            # else:
                            return self.eps

                    e_obj = Expression(self.flowsheet().config.time,
                                       self.config.property_package.phase_list,
                                       rule=e_rule)

                elif l_name.endswith("_comp"):
                    if self.config.split_basis == \
                            SplittingType.componentFlow:
                        def e_rule(b, t, j):
                            if split_map[j] == o:
                                return mb[t].component(l_name)[j]
                            else:
                                return self.eps

                    else:
                        def e_rule(b, t, j):
                            mfp = mb[t].component(
                                        "{0}_phase{1}"
                                        .format(l_name[:-5], l_name[-5:]))

                            if mfp is None:
                                raise AttributeError(
                                    "{} Cannot use ideal splitting with this "
                                    "property package. Package uses indexed "
                                    "port member {} which does not have the "
                                    "correct indexing sets, and an equivalent "
                                    "variable with correct indexing sets is "
                                    "not available."
                                    .format(self.name, s))

                            for p in self.config.property_package.phase_list:
                                if self.config.split_basis == \
                                        SplittingType.phaseFlow:
                                    s_check = split_map[p]
                                elif self.config.split_basis == \
                                        SplittingType.phaseComponentFlow:
                                    s_check = split_map[p, j]
                                else:
                                    raise BurntToast(
                                        "{} This should not happen. Please"
                                        " report this bug to the IDAES "
                                        "developers.".format(self.name))

                                if s_check == o:
                                    return mfp[p, j]
                            # else:
                            return self.eps

                    e_obj = Expression(
                                self.flowsheet().config.time,
                                self.config.property_package.component_list,
                                rule=e_rule)

                else:
                    def e_rule(b, t):
                        try:
                            if self.config.split_basis == \
                                    SplittingType.phaseFlow:
                                ivar = mb[t].component(l_name+"_phase")
                                if ivar is not None:
                                    for p in self.config \
                                            .property_package.phase_list:
                                        if split_map[p] == o:
                                            return ivar[p]
                                        else:
                                            continue
                                else:
                                    ivar = mb[t].component(
                                            l_name+"_phase_comp")
                                    if ivar is not None:
                                        for p in self.config \
                                                .property_package.phase_list:
                                            if split_map[p] == o:
                                                return sum(
                                                    ivar[p, j] for j in
                                                    self.config
                                                    .property_package
                                                    .component_list)
                                            else:
                                                continue
                                    else:
                                        raise AttributeError

                            elif self.config.split_basis == \
                                    SplittingType.componentFlow:
                                ivar = mb[t].component(l_name+"_comp")
                                if ivar is not None:
                                    for j in self.config.property_package \
                                            .component_list:
                                        if split_map[j] == o:
                                            return ivar[j]
                                        else:
                                            continue
                                else:
                                    ivar = mb[t].component(l_name +
                                                           "_phase_comp")
                                    if ivar is not None:
                                        for j in self.config.property_package \
                                                .component_list:
                                            if split_map[j] == o:
                                                return sum(
                                                    ivar[p, j] for p in
                                                    self.config
                                                    .property_package
                                                    .phase_list)
                                            else:
                                                continue
                                    else:
                                        raise AttributeError
                            elif self.config.split_basis == \
                                    SplittingType.phaseComponentFlow:
                                ivar = mb[t].component(l_name+"_phase_comp")
                                if ivar is not None:
                                    for p in self.config.property_package \
                                            .phase_list:
                                        for j in self.config.property_package \
                                                .component_list:
                                            if split_map[p, j] == o:
                                                return ivar[p, j]
                                            else:
                                                continue
                                else:
                                    raise AttributeError
                            else:
                                # Unrecognised split tupe
                                raise BurntToast(
                                    "{} received unrecognised value for "
                                    "split_basis argumnet. This should never "
                                    "happen, so please contact the IDAES "
                                    "developers with this bug."
                                    .format(self.name))

                        except:
                            # If cannot find equivalent var, raise exception
                            raise AttributeError(
                                "{} Cannot use ideal splitting with this "
                                "property package. Package uses unindexed "
                                "port member {} which does not have an "
                                "equivalent indexed form."
                                .format(self.name, s))

                    e_obj = Expression(self.flowsheet().config.time,
                                       rule=e_rule)

                # Add Reference/Expression object to Separator model object
                setattr(self, "_"+o+"_"+l_name+"_ref", e_obj)

                # Add member to Port object
                p_obj.add(e_obj, s)

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().config.time:
            try:
                if blk.config.mixed_state_block is None:
                    blk.mixed_state[t].model_check()
                else:
                    blk.config.mixed_state_block.model_check()
            except AttributeError:
                _log.warning('{} Separator inlet state block has no '
                             'model check. To correct this, add a '
                             'model_check method to the associated '
                             'StateBlock class.'.format(blk.name))

            try:
                outlet_list = blk.create_outlet_list()
                for o in outlet_list:
                    o_block = getattr(blk, o+"_state")
                    o_block[t].model_check()
            except AttributeError:
                _log.warning('{} Separator outlet state block has no '
                             'model checks. To correct this, add a model_check'
                             ' method to the associated StateBlock class.'
                             .format(blk.name))

    def initialize(blk, outlvl=6, optarg={},
                   solver='ipopt', hold_state=False):
        '''
        Initialization routine for separator (default solver ipopt)

        Keyword Arguments:
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        '''
        init_log = getInitLogger(blk.name, outlvl)
        # Set solver options
        opt = SolverFactory(solver)
        opt.options = optarg

        # Initialize mixed state block
        if blk.config.mixed_state_block is not None:
            mblock = blk.config.mixed_state_block
        else:
            mblock = blk.mixed_state
        flags = mblock.initialize(outlvl=outlvl+1,
                                  optarg=optarg,
                                  solver=solver,
                                  hold_state=True)

        if blk.config.ideal_separation:
            # If using ideal splitting, initialization should be complete
            return flags

        # Initialize outlet StateBlocks
        outlet_list = blk.create_outlet_list()

        for o in outlet_list:
            # Get corresponding outlet StateBlock
            o_block = getattr(blk, o+"_state")

            # Create dict to store fixed status of state variables
            o_flags = {}
            for t in blk.flowsheet().config.time:

                # Calculate values for state variables
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        # Record whether variable was fixed or not
                        o_flags[t, v, k] = s_vars[v][k].fixed

                        # If fixed, use current value
                        # otherwise calculate guess from mixed state and fix
                        if not s_vars[v][k].fixed:
                            m_var = getattr(mblock[t], s_vars[v].local_name)
                            if "flow" in v:
                                # If a "flow" variable, is extensive
                                # Apply split fraction
                                try:
                                    if (k is None or blk.config.split_basis ==
                                            SplittingType.totalFlow):
                                        s_vars[v][k].fix(value(
                                            m_var[k] *
                                            blk.split_fraction[(t, o)]))
                                    else:
                                        s_vars[v][k].fix(value(
                                                m_var[k]*blk.split_fraction[
                                                        (t, o) + k]))
                                except KeyError:
                                    raise KeyError(
                                        "{} state variable and split fraction "
                                        "indexing sets do not match. The "
                                        "in-built  initialization routine for "
                                        "Separators relies on the split "
                                        "fraction and state variable indexing "
                                        "sets matching to calculate initial "
                                        "guesses for extensive variables. In "
                                        "other cases users will need to "
                                        "provide their own initial guesses"
                                        .format(blk.name))
                            else:
                                # Otherwise intensive, equate to mixed stream
                                s_vars[v][k].fix(m_var[k].value)

            # Call initialization routine for outlet StateBlock
            o_block.initialize(outlvl=outlvl+1,
                               optarg=optarg,
                               solver=solver,
                               hold_state=False)

            # Revert fixed status of variables to what they were before
            for t in blk.flowsheet().config.time:
                for v in s_vars:
                    for k in s_vars[v]:
                        s_vars[v][k].fixed = o_flags[t, v, k]

        if blk.config.mixed_state_block is None:
            results = opt.solve(blk, tee=init_tee(init_log))
            init_log.log(5, "Initialization Complete: {}".format(condition(results)))
        else:
            init_log.log(5, "Initialization Complete.")

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags, outlvl=outlvl+1)

    def release_state(blk, flags, outlvl=6):
        '''
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        '''
        if blk.config.mixed_state_block is None:
            mblock = blk.mixed_state
        else:
            mblock = blk.config.mixed_state_block

        mblock.release_state(flags, outlvl=outlvl+1)

    def _get_performance_contents(self, time_point=0):
        if hasattr(self, "split_fraction"):
            for k in self.split_fraction.keys():
                if k[0] == time_point:
                    var_dict = {f"Split Fraction [{str(k[1:])}]":
                                self.split_fraction[k]}
            return {"vars": var_dict}
        else:
            return {}

    def _get_stream_table_contents(self, time_point=0):
        outlet_list = self.create_outlet_list()

        if not self.config.ideal_separation:
            io_dict = {}
            if self.config.mixed_state_block is None:
                io_dict["Inlet"] = self.mixed_state
            else:
                io_dict["Inlet"] = self.config.mixed_state_block

                for o in outlet_list:
                    io_dict[o] = getattr(self, o+"_state")
            return create_stream_table_dataframe(
                    io_dict,
                    time_point=time_point)

        else:
            stream_attributes = {}

            for n in outlet_list+["inlet"]:
                port_obj = getattr(self, n)

                stream_attributes[n] = {}

                for k in port_obj.vars:
                    for i in port_obj.vars[k]:
                        if isinstance(i, float):
                            stream_attributes[n][k] = value(
                                port_obj.vars[k][time_point])
                        else:
                            if len(i) == 2:
                                kname = str(i[1])
                            else:
                                kname = str(i[1:])
                            stream_attributes[n][k+" "+kname] = \
                                value(port_obj.vars[k][time_point, i[1:]])

            return DataFrame.from_dict(stream_attributes, orient="columns")



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Generic template for a translator block.
"""
# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

__author__ = "Andrew Lee"


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("Translator")
class TranslatorData(UnitModelBlockData):
    """
    Standard Translator Block Class
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Translator blocks are always steady-state."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Translator blocks do not contain holdup."""))
    CONFIG.declare("outlet_state_defined", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Indicated whether outlet state will be fully defined",
        doc="""Indicates whether unit model will fully define outlet state.
If False, the outlet property package will enforce constraints such as sum
of mole fractions and phase equilibrium.
**default** - True.
**Valid values:** {
**True** - outlet state will be fully defined,
**False** - outlet property package should enforce sumation and equilibrium
constraints.}"""))
    CONFIG.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Indicated whether outlet state is in phase equilibrium",
        doc="""Indicates whether outlet property package should enforce
phase equilibrium constraints.
**default** - False.
**Valid values:** {
**True** - outlet property package should calculate phase equilibrium,
**False** - outlet property package should notcalculate phase equilibrium.}
"""))
    CONFIG.declare("inlet_property_package", ConfigValue(
        default=None,
        domain=is_physical_parameter_block,
        description="Property package to use for incoming stream",
        doc="""Property parameter object used to define property calculations
for the incoming stream,
**default** - None.
**Valid values:** {
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("inlet_property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property package of "
                    "the incoming stream",
        doc="""A ConfigBlock with arguments to be passed to the property block
associated with the incoming stream,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("outlet_property_package", ConfigValue(
        default=None,
        domain=is_physical_parameter_block,
        description="Property package to use for outgoing stream",
        doc="""Property parameter object used to define property calculations
for the outgoing stream,
**default** - None.
**Valid values:** {
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("outlet_property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property package of "
                    "the outgoing stream",
        doc="""A ConfigBlock with arguments to be passed to the property block
associated with the outgoing stream,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TranslatorData, self).build()

        # Check construction argumnet consistency
        if (self.config.outlet_state_defined and
                self.config.has_phase_equilibrium):
            raise ConfigurationError(
                "{} cannot calcuate phase equilibrium (has_phase_equilibrium "
                "= True) when outlet state is set to be fully defined ("
                "outlet_state_defined = True).".format(self.name))

        # Add State Blocks
        self.properties_in = (
                    self.config.inlet_property_package.state_block_class(
                        self.flowsheet().config.time,
                        doc="Material properties in incoming stream",
                        default={
                            "defined_state": True,
                            "parameters": self.config.inlet_property_package,
                            "has_phase_equilibrium": False,
                            **self.config.inlet_property_package_args}))

        self.properties_out = (
                    self.config.outlet_property_package.state_block_class(
                        self.flowsheet().config.time,
                        doc="Material properties in outgoing stream",
                        default={
                            "defined_state": self.config.outlet_state_defined,
                            "parameters": self.config.outlet_property_package,
                            "has_phase_equilibrium":
                                self.config.has_phase_equilibrium,
                            **self.config.outlet_property_package_args}))

        # Add outlet port
        self.add_port(name="inlet",
                      block=self.properties_in,
                      doc="Inlet Port")
        self.add_port(name="outlet",
                      block=self.properties_out,
                      doc="Outlet Port")

    def initialize(blk, state_args_in={}, state_args_out={}, outlvl=6,
                   solver='ipopt', optarg={'tol': 1e-6}):
        '''
        This method calls the initialization method of the state blocks.

        Keyword Arguments:
            state_args_in : a dict of arguments to be passed to the inlet
                            property package (to provide an initial state for
                            initialization (see documentation of the specific
                            property package) (default = {}).
            state_args_out : a dict of arguments to be passed to the outlet
                             property package (to provide an initial state for
                             initialization (see documentation of the specific
                             property package) (default = {}).
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating which solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        '''
        init_log = getInitLogger(blk.name, outlvl)
        # ---------------------------------------------------------------------
        # Initialize state block
        flags = blk.properties_in.initialize(outlvl=outlvl+1,
                                             optarg=optarg,
                                             solver=solver,
                                             state_args=state_args_in,
                                             hold_state=True)

        blk.properties_out.initialize(outlvl=outlvl+1,
                                      optarg=optarg,
                                      solver=solver,
                                      state_args=state_args_out)

        blk.properties_in.release_state(flags)

        init_log.log(5, 'Initialization Complete.')
